/*
 * Copyright (c) 2009 The openGion Project.
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *     http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND,
 * either express or implied. See the License for the specific language
 * governing permissions and limitations under the License.
 */
package org.opengion.fukurou.process;

import org.opengion.fukurou.util.Argument;
import org.opengion.fukurou.util.FileUtil;
import org.opengion.fukurou.util.FileString;
import org.opengion.fukurou.util.Closer ;
import org.opengion.fukurou.util.StringUtil ;
import org.opengion.fukurou.util.LogWriter;

import java.util.Arrays;
import java.util.Map ;
import java.util.LinkedHashMap ;
import java.util.regex.Pattern;
import java.util.regex.Matcher;

import java.io.File;
import java.io.PrintWriter;
import java.io.BufferedReader;
import java.io.IOException;

/**
 * Process_Grep は、上流から受け取った FileLineModelから、文字列を見つけ出す
 * ChainProcess インターフェースの実装クラスです。
 *
 * 正規表現の keyword を上流から受け取った FileLineModel から検索します。
 * 見つかった対象ファイルから、指定の文字列を置換する場合は、-change か
 * -changeFile で、keyword を置換する文字列を指定して下さい。
 * 置換する文字列には、\t と \n の特殊文字が使用できます。
 *
 * 処理対象は、通常は、１行づつ読み取りながら処理を行います。存在チェックの場合は、
 * 見つかった時点で処理を中止します。これは、該当箇所をピックアップするのではなく、
 * 存在しているかどうかを判断して、あれば、下流に流すというのが目的だからです。
 * keyword を、改行を含む正規表現で、検索・置換する場合は、-useBulkRead 属性を
 * true に設定してください。これは、入力ファイルを一括して読み込みます。
 * -ignoreCase は、正規表現の検索時にキーの大文字小文字を無視するように指定します。
 * -notEquals は、結果(見つかればtrue)を反転(見つからなければtrue)します。
 * これは、行単位ではなく、ファイル単位に判定しますので、change 指定した場合
 * でも、対象行は、見つかった行です。ただし、下流に対して、見つからない
 * 場合だけ処理を継続させます。
 * -inEncode は、入力ファイルのエンコード指定になります。
 * -outEncode は、出力ファイルのエンコードや、changeFileで指定の置換文字列ファイルの
 * エンコード指定になります。(changeFile は、必ず 出力ファイルと同じエンコードです。)
 * これらのエンコードが無指定の場合は、System.getProperty("file.encoding") で
 * 求まる値を使用します。
 * -changeFile を使用することで、複数行の文字列に置換することが可能です。
 * -outfile では、処理を行ったファイル名一覧をセーブします。
 *
 * 上流(プロセスチェインのデータは上流から渡されます。)からのLineModel の
 * ファイルオブジェクトより、指定の文字列が含まれているか検索します。
 * 上流プロセスでは、Name 属性として、『File』を持ち、値は、Fileオブジェクト
 * である、Process_FileSearch を使用するのが、便利です。それ以外のクラスを
 * 使用する場合でも、Name属性と、File オブジェクトを持つ LineModel を受け渡し
 * できれば、使用可能です。
 *
 * 引数文字列中に空白を含む場合は、ダブルコーテーション("") で括って下さい。
 * 引数文字列の 『=』の前後には、空白は挟めません。必ず、-key=value の様に
 * 繋げてください。
 *
 * @og.formSample
 *  Process_Grep -keyword=検索文字列 -ignoreCase=true -outfile=OUTFILE -encode=UTF-8
 *
 *    -keyword=キーワード        ：検索する語句
 *   [-ignoreCase=大文字小文字 ] ：検索時に大文字小文字を区別しない(true)かどうか(初期値:区別する[false])
 *   [-notEquals=判定結果の反転] ：判定結果を反転させる(true)かどうか(初期値:反転させない[false])
 *   [-inEncode=入力エンコード ] ：入力ファイルのエンコードタイプ
 *   [-outEncode=出力エンコード] ：出力ファイルや置換ファイルのエンコードタイプ
 *   [-change=置換文字列       ] ：-change="ABCD" \t や \n などの特殊文字が使用できます。
 *   [-changeFile=置換ファイル ] ：-changeFile=change.txt このファイルの記述すべてと置換します。
 *                                     -change と、-changeFile は、同時に指定できません。
 *                                     置換機能使用時は、必ず、_backup というファイルが作成されます。
 *   [-insert=[CHANGE/BEFORE/AFTER]   ] : 置換でなく挿入する場合の位置を指定します(初期値:CHANGE)
 *                                 スペースで区切って数字を記述すると、挿入位置にオフセットできます。
 *   [-delete=[false/true]     ] : 置換でなく削除します(初期値:false)
 *   [-useBackup=[false/true]  ] ：trueは、backupファイルを作成します(初期値:false)
 *   [-useBulkRead=[false/true]] ：trueは、入力ファイルを一括読込します(初期値:false)
 *   [-display=[false/true]    ] ：trueは、検索状況を表示します(初期値:false)
 *   [-debug=[false/true]      ] ：デバッグ情報を標準出力に表示する(true)かしない(false)か(初期値:false[表示しない])
 *
 * @version  4.0
 * @author   Kazuhiko Hasegawa
 * @since    JDK5.0,
 */
public class Process_Grep extends AbstractProcess implements ChainProcess {
	private static final String[] INSERT_LIST = new String[] { "CHANGE","BEFORE","AFTER" };

	private Pattern pattern		= null;
	private String	keyword		= null;
	private boolean	ignoreCase	= false;
	private boolean	notEquals	= false;
	private String	inEncode	= null;
	private String	outEncode	= null;
	private String	change		= null;
	private String	insert		= "CHANGE";		// "CHANGE","BEFORE","AFTER" のどれか
	private int		insOffset	= 0;			// "BEFORE","AFTER" 時のオフセット
	private boolean	useBackup	= false;
	private boolean	useBulkRead	= false;		// 4.0.1.0 (2007/12/14)
	private boolean	delete		= false;
	private boolean	display		= false;
	private boolean	debug		= false;		// 5.1.2.0 (2010/01/01)

	private int		inCount		= 0;
	private int		findCount	= 0;
	private int		cngCount	= 0;

	private static final Map<String,String> mustProparty   ;		// ［プロパティ］必須チェック用 Map
	private static final Map<String,String> usableProparty ;		// ［プロパティ］整合性チェック Map

	static {
		mustProparty = new LinkedHashMap<String,String>();
		mustProparty.put( "keyword",	"検索する語句(必須)" );

		usableProparty = new LinkedHashMap<String,String>();
		usableProparty.put( "ignoreCase",	"検索時に大文字小文字を区別しない(true)かどうか。" +
										CR + "(初期値:区別する[false])" );
		usableProparty.put( "notEquals",	"検索時に判定結果を反転させる(true)かどうか。" +
										CR + "(初期値:反転させない[false])" );
		usableProparty.put( "inEncode",		"入力ファイルのエンコードタイプ" );
		usableProparty.put( "outEncode",	"出力ファイルや置換ファイルのエンコードタイプ" );
		usableProparty.put( "change",		"置換文字列 例: -change=\"ABCD\" \\t や \\n などの特殊文字が使用できます。" );
		usableProparty.put( "changeFile",	"置換文字列ファイル 例: -changeFile=change.txt" +
										CR + "-change と、-changeFile は、同時に指定できません。" +
										CR + "置換機能使用時は、必ず、_backup というファイルが作成されます。" );
		usableProparty.put( "insert",		"[CHANGE/BEFORE/AFTER]:置換でなく挿入する場合の位置を指定します(初期値:CHANGE)"  +
										CR + "スペースで区切って数字を記述すると、挿入位置にオフセットできます。" );
		usableProparty.put( "delete",		"[false/true]:trueは、置換でなく削除します(初期値:false)" );
		usableProparty.put( "useBackup",	"[false/true]:trueは、backupファイルを作成します(初期値:false)" );
		usableProparty.put( "useBulkRead",	"[false/true]:trueは、入力ファイルを一括読込します(初期値:false)" );
		usableProparty.put( "display",		"[false/true]:trueは、検索状況を表示します(初期値:false)" );
		usableProparty.put( "debug",		"デバッグ情報を標準出力に表示する(true)かしない(false)か" +
											CR + "(初期値:false:表示しない)" );
	}

	/**
	 * デフォルトコンストラクター。
	 * このクラスは、動的作成されます。デフォルトコンストラクターで、
	 * super クラスに対して、必要な初期化を行っておきます。
	 *
	 */
	public Process_Grep() {
		super( "org.opengion.fukurou.process.Process_Grep",mustProparty,usableProparty );
	}

	/**
	 * プロセスの初期化を行います。初めに一度だけ、呼び出されます。
	 * 初期処理(ファイルオープン、ＤＢオープン等)に使用します。
	 *
	 * @param   paramProcess データベースの接続先情報などを持っているオブジェクト
	 */
	public void init( final ParamProcess paramProcess ) {
		Argument arg = getArgument();

		keyword			= arg.getProparty("keyword");
		ignoreCase		= arg.getProparty("ignoreCase"	,ignoreCase);
		notEquals		= arg.getProparty("notEquals"	,notEquals);
		inEncode		= arg.getProparty("inEncode"	,System.getProperty("file.encoding"));
		outEncode		= arg.getProparty("outEncode"	,System.getProperty("file.encoding"));
		useBackup		= arg.getProparty("useBackup"	,useBackup);
		useBulkRead		= arg.getProparty("useBulkRead"	,useBulkRead);	// 4.0.1.0 (2007/12/14)
		delete			= arg.getProparty("delete"		,delete );
		insert	  		= arg.getProparty("insert" ,insert );
		change			= arg.getFileProparty( "change","changeFile",outEncode,false );
		display			= arg.getProparty("display"		,display);
		debug			= arg.getProparty( "debug"  ,debug );		// 5.1.2.0 (2010/01/01)

		if( change != null ) {
			int adrs = insert.indexOf( ' ' );	// オフセット数字の有無
			if( adrs > 0 ) {
				insOffset = Integer.parseInt( insert.substring( adrs+1 ) );
				insert    = insert.substring( 0,adrs );
			}

			boolean isOK = false;
			for( int i=0; i<INSERT_LIST.length; i++ ) {
				if( insert.equalsIgnoreCase( INSERT_LIST[i] ) ) {
					isOK = true; break;
				}
			}
			if( !isOK ) {
				String errMsg = "insert は、" + Arrays.toString( INSERT_LIST )
									+ " から指定してください。" + CR
									+ "-insert=[" + insert + "]" ;
				throw new RuntimeException( errMsg );
			}

			change = StringUtil.replace( change,"\\n",CR );
			change = StringUtil.replace( change,"\\t","\t" );
		}

		if( delete ) { change = ""; }	// 削除は、"" 文字列と置換します。

		if( ignoreCase ) {
			pattern = Pattern.compile( keyword,Pattern.CASE_INSENSITIVE );
		}
		else {
			pattern = Pattern.compile( keyword );
		}
	}

	/**
	 * プロセスの終了を行います。最後に一度だけ、呼び出されます。
	 * 終了処理(ファイルクローズ、ＤＢクローズ等)に使用します。
	 *
	 * @param   isOK トータルで、OKだったかどうか[true:成功/false:失敗]
	 */
	public void end( final boolean isOK ) {
		// ここでは処理を行いません。
	}

	/**
	 * 引数の LineModel を処理するメソッドです。
	 * 変換処理後の LineModel を返します。
	 * 後続処理を行わない場合(データのフィルタリングを行う場合)は、
	 * null データを返します。つまり、null データは、後続処理を行わない
	 * フラグの代わりにも使用しています。
	 * なお、変換処理後の LineModel と、オリジナルの LineModel が、
	 * 同一か、コピー(クローン)かは、各処理メソッド内で決めています。
	 * ドキュメントに明記されていない場合は、副作用が問題になる場合は、
	 * 各処理ごとに自分でコピー(クローン)して下さい。
	 *
	 * @og.rev 4.0.1.0 (2007/12/14) ファイルの一括処理対応。
	 * @og.rev 5.7.2.2 (2014/01/24) エラー時にデータも出力します。
	 *
	 * @param	data	オリジナルのLineModel
	 *
	 * @return	処理変換後のLineModel
	 */
	public LineModel action( final LineModel data ) {
		inCount++ ;

		final FileLineModel fileData ;
		if( data instanceof FileLineModel ) {
			fileData = (FileLineModel)data ;
		}
		else {
			String errMsg = "データが FileLineModel オブジェクトではありません。" + CR ;
			throw new RuntimeException( errMsg );
		}

		File file = fileData.getFile() ;
		if( ! file.isFile() ) {
			if( display ) { println( data.dataLine() ); }		// 5.1.2.0 (2010/01/01) display の条件変更
			return data;
		}

		final boolean isFind ;
		try {
			String fileLine = null;
			int firstLineNo = -1;
			if( useBulkRead ) { fileLine    = findKeywordAsBulk( file ); }
			else			  { firstLineNo = findKeyword( file ); }

			isFind = fileLine != null || firstLineNo >= 0 ;

			// 置換処理 ただし、見つかったときのみ実行
			if( change != null && isFind ) {
				// 入力ファイルは、オリジナル_backup ファイルとする。過去のファイルを削除
				File inFile = new File( file.getPath() + "_backup" );
				if( inFile.exists() && !inFile.delete() ) {
					String errMsg = "過去のBKUPファイルを削除できませんでした。[" + inFile + "]" + CR
								+	"data=[" + data.dataLine() + "]" + CR ;		// 5.7.2.2 (2014/01/24) エラー時にデータも出力します。
					throw new RuntimeException( errMsg );
				}

				// オリジナルのファイルを、_backup ファイル名に先に変換する。
				File fromFile = new File( file.getPath() );
				if( !fromFile.renameTo( inFile ) ) {
					String errMsg = "所定のファイルをリネームできませんでした。[" + fromFile + "]" + CR
								+	"data=[" + data.dataLine() + "]" + CR ;		// 5.7.2.2 (2014/01/24) エラー時にデータも出力します。
					throw new RuntimeException( errMsg );
				}

				// 変換処理 本体
				if( useBulkRead ) { changeKeywordAsBulk( fileLine,file ); }
				else			  { changeKeyword( inFile,file,firstLineNo ); }

				// backup を使わない場合は、削除する。
				// 4.0.0.0 (2007/11/29) 入れ子if の統合
				if( ! useBackup && !inFile.delete() ) {
					String errMsg = "所定のファイルを削除できませんでした。[" + inFile + "]" + CR
								+	"data=[" + data.dataLine() + "]" + CR ;		// 5.7.2.2 (2014/01/24) エラー時にデータも出力します。
					throw new RuntimeException( errMsg );
				}
			}
		}
		catch ( RuntimeException ex ) {
			String errMsg = "処理中にエラーが発生しました。[" + data.getRowNo() + "]件目" + CR
						+	"data=[" + data.dataLine() + "]" + CR ;		// 5.7.2.2 (2014/01/24) エラー時にデータも出力します。
			throw new RuntimeException( errMsg,ex );
		}

		if( display && ( notEquals ^ isFind ) ) { println( data.dataLine() ); }		// 5.1.2.0 (2010/01/01) display の条件変更
		return ( notEquals ^ isFind ) ? data : null ;
	}

	/**
	 * キーワードが存在しているかどうかをチェックします。
	 * ここでは、１行づつ読み取りながら、最初に見つかった時点で制御を返します。
	 * よって、複数行にまたがる keyword でのマッチングは出来ませんが、大きな
	 * ファイル等での検索には、効率的です。
	 *
	 * @og.rev 4.0.1.0 (2007/12/14) 新規追加
	 *
	 * @param	file	検索元のファイルオブジェクト
	 *
	 * @return	最初に見つかった行番号(見つからなければ、-1 を返す)
	 */
	private int findKeyword( final File file ) {
		BufferedReader reader = null;

		int firstLineNo = -1;
		try {
			reader = FileUtil.getBufferedReader( file,inEncode );
			String line ;
			int lineNo = 0;
			while((line = reader.readLine()) != null) {
				lineNo++ ;
				Matcher mach = pattern.matcher( line );
				if( mach.find() ) {
					if( debug ) {
						String buf = "DEBUG:\t" + file.getPath() + "(" + lineNo + "): " + line ;
						println( buf );
					}
					firstLineNo = lineNo;
					break;
				}
			}
		}
		catch ( IOException ex ) {
			String errMsg = "ファイル読取エラーが発生しました。[" + file.getPath() + "]" ;
			throw new RuntimeException( errMsg,ex );
		}
		finally {
			Closer.ioClose( reader );
		}

		return firstLineNo;
	}

	/**
	 * キーワードが存在しているかどうかをチェックします。
	 * ここでは、ファイルをすべて読み取ってから、チェックします。
	 * よって、複数行にまたがる keyword でのマッチングが可能です。
	 *
	 * @og.rev 4.0.1.0 (2007/12/14) 新規追加
	 *
	 * @param	file	検索元のファイルオブジェクト
	 *
	 * @return	検索元のファイルの文字列化情報(ただし、見つからなければ、null)
	 */
	private String findKeywordAsBulk( final File file ) {

		boolean isFind = false;

		FileString sf = new FileString();
		sf.setFilename( file.getPath() );
		sf.setEncode( inEncode );
		String line = sf.getValue();

		Matcher mach = pattern.matcher( line );
		if( mach.find() ) {
			if( debug ) { println( "DEBUG:\t" + file.getPath() ); }
			isFind = true;
		}

		return isFind ? line : null;
	}

	/**
	 * キーワードを指定の文字列に置き換えます。
	 * useBackup 属性に true を指定した場合、置き換え後の、backup ファイルは、
	 * オリジナル_backup という名称に変わります。
	 * ここでは、１行づつ読み取りながら、変換処理を行います。
	 * よって、複数行にまたがる keyword でのマッチングは出来ませんが、大きな
	 * ファイル等での置換でも、メモリの使用量は抑えられます。
	 *
	 * @og.rev 4.0.1.0 (2007/12/14) 置換処理を独立させます。
	 *
	 * @param	inFile	検索元の入力ファイルオブジェクト
	 * @param	outFile	変換後の出力ファイルオブジェクト
	 * @param	firstLineNo	キーワードが存在した場合の最初の行番号
	 */
	private void changeKeyword( final File inFile,final File outFile,final int firstLineNo ) {

		BufferedReader reader = FileUtil.getBufferedReader( inFile,inEncode );
		PrintWriter    writer = FileUtil.getPrintWriter( outFile,outEncode );

		String line = null;
		try {
			int lineNo = 0;
			while((line = reader.readLine()) != null) {
				lineNo++ ;
				if( lineNo >= firstLineNo ) {
					Matcher mach = pattern.matcher( line );

					String chnStr = null;
					if( "CHANGE".equals( insert ) ) {
						chnStr = strChange( mach );
					}
					else if( "BEFORE".equals( insert ) ) {
						chnStr = strBefore( line,mach );
					}
					else if( "AFTER".equals( insert ) ) {
						chnStr = strAfter( line,mach );
					}

					if( chnStr != null ) {
						line = chnStr;
						cngCount++ ;	// 変換されれば カウント
					}
				}
				writer.println( line );	// readLine() してるので、最後に改行が必要。
			}
		}
		catch ( IOException ex ) {
			String errMsg = "処理中にエラーが発生しました。[" + line + "]" ;
			throw new RuntimeException( errMsg,ex );
		}
		finally {
			Closer.ioClose( reader );
			Closer.ioClose( writer );
		}
	}
	/**
	 * キーワードを指定の文字列に置き換えます。
	 * useBackup 属性に true を指定した場合、置き換え後の、backup ファイルは、
	 * オリジナル_backup という名称に変わります。
	 * ここでは、ファイルをすべて読み取ってから、チェックします。
	 * よって、複数行にまたがる keyword でのマッチングが可能です。
	 *
	 * @og.rev 4.0.1.0 (2007/12/14) 置換処理を独立させます。
	 *
	 * @param	fileLine	検索元の行文字列
	 * @param	outFile	出力ファイルオブジェクト
	 */
	private void changeKeywordAsBulk( final String fileLine,final File outFile ) {
		PrintWriter writer = FileUtil.getPrintWriter( outFile,outEncode );

		String line = fileLine ;
		try {
			Matcher mach = pattern.matcher( line );

			String chnStr = null;
			if( "CHANGE".equals( insert ) ) {
				chnStr = strChange( mach );
			}
			else if( "BEFORE".equals( insert ) ) {
				chnStr = strBefore( line,mach );
			}
			else if( "AFTER".equals( insert ) ) {
				chnStr = strAfter( line,mach );
			}

			if( chnStr != null ) {
				line = chnStr;
				cngCount++ ;	// 変換されれば カウント
			}

			writer.print( line );	// 注意：改行コードは、不要
		}
		catch ( RuntimeException ex ) {
			String errMsg = "処理中にエラーが発生しました。[" + outFile.getPath() + "]" ;
			throw new RuntimeException( errMsg,ex );
		}
		finally {
			Closer.ioClose( writer );
		}
	}

	/**
	 * insert が、"CHANGE" の場合の処理結果を求めます。
	 * 変換しなかった場合は、null を返します。
	 * これは、変換カウントを算出する為のフラグ代わりに使用しています。
	 *
	 * @param	mach	キーワードの正規表現
	 *
	 * @return	変換結果(対象行で無い場合は、null)
	 */
	private String strChange( final Matcher mach ) {
		String line = null;
		if( mach.find() ) {
			line = mach.replaceAll( change );
		}
		return line ;
	}

	/**
	 * insert が、"BEFORE" の場合の処理結果を求めます。
	 * 変換しなかった場合は、null を返します。
	 * これは、変換カウントを算出する為のフラグ代わりに使用しています。
	 *
	 * @param	line	検索行
	 * @param	mach	キーワードの正規表現
	 *
	 * @return	変換結果(対象行で無い場合は、null)
	 */
	private String strBefore( final String line , final Matcher mach ) {
		boolean isChng = false;
		StringBuilder buf = new StringBuilder( line.length() );
		int indx = 0;
		while( mach.find() ) {
			isChng = true;
			int strt = mach.start() + insOffset;
			buf.append( line.substring( indx,strt ) );
			buf.append( change );
			indx = strt;
		}

		String rtn = null;
		if( isChng ) {
			buf.append( line.substring( indx ) );
			rtn = buf.toString();
		}

		return rtn ;
	}

	/**
	 * insert が、"AFTER" の場合の処理結果を求めます。
	 * 変換しなかった場合は、null を返します。
	 * これは、変換カウントを算出する為のフラグ代わりに使用しています。
	 *
	 * @param	line	検索行
	 * @param	mach	キーワードの正規表現
	 *
	 * @return	変換結果(対象行で無い場合は、null)
	 */
	private String strAfter( final String line , final Matcher mach ) {
		boolean isChng = false;
		StringBuilder buf = new StringBuilder( line.length() );
		int indx = 0;
		while( mach.find() ) {
			isChng = true;
			int end = mach.end() + insOffset;
			buf.append( line.substring( indx,end ) );
			buf.append( change );
			indx = end;
		}
		String rtn = null;
		if( isChng ) {
			buf.append( line.substring( indx ) );
			rtn = buf.toString();
		}

		return rtn ;
	}

	/**
	 * プロセスの処理結果のレポート表現を返します。
	 * 処理プログラム名、入力件数、出力件数などの情報です。
	 * この文字列をそのまま、標準出力に出すことで、結果レポートと出来るような
	 * 形式で出してください。
	 *
	 * @return   処理結果のレポート
	 */
	public String report() {
		if( findCount < cngCount ) { findCount = cngCount; }

		String report = "[" + getClass().getName() + "]" + CR
				+ TAB + "Search Keyword    : " + keyword    + CR
				+ TAB + "Search File Count : " + inCount    + CR
				+ TAB + "Key Find    Count : " + findCount  + CR
				+ TAB + "Key Change  Count : " + cngCount ;

		return report ;
	}

	/**
	 * このクラスの使用方法を返します。
	 *
	 * @return	このクラスの使用方法
	 */
	public String usage() {
		StringBuilder buf = new StringBuilder();

		buf.append( "Process_Grep は、上流から受け取った FileLineModelから、文字列を見つけ出す"		).append( CR );
		buf.append( "ChainProcess インターフェースの実装クラスです。"								).append( CR );
		buf.append( CR );
		buf.append( "正規表現の keyword を上流から受け取った FileLineModel から検索します。"		).append( CR );
		buf.append( "見つかった対象ファイルから、指定の文字列を置換する場合は、-change か" 			).append( CR );
		buf.append( "-changeFile で、keyword を置換する文字列を指定して下さい。" 					).append( CR );
		buf.append( "置換する文字列には、\t と \n の特殊文字が使用できます。" 						).append( CR );
		buf.append( CR );
		buf.append( "処理対象は、通常は、１行づつ読み取りながら処理を行います。存在チェックの場合は、"	).append( CR );
		buf.append( "見つかった時点で処理を中止します。これは、該当箇所をピックアップするのではなく、"	).append( CR );
		buf.append( "存在しているかどうかを判断して、あれば、下流に流すというのが目的だからです。"		).append( CR );
		buf.append( "keyword を、改行を含む正規表現で、検索・置換する場合は、-useBulkRead 属性を"		).append( CR );
		buf.append( "true に設定してください。これは、入力ファイルを一括して読み込みます。"			).append( CR );
		buf.append( "-ignoreCase は、検索時にキーの大文字小文字を無視するように指定します。"		).append( CR );
		buf.append( "-notEquals は、結果(見つかればtrue)を反転(見つからなければtrue)します。"		).append( CR );
		buf.append( "これは、行単位ではなく、ファイル単位に判定しますので、change 指定した場合"		).append( CR );
		buf.append( "でも、対象行は、見つかった行です。ただし、下流に対して、見つからない"			).append( CR );
		buf.append( "場合だけ処理を継続させます。"													).append( CR );
		buf.append( "-inEncode は、入力ファイルのエンコード指定になります。" 						).append( CR );
		buf.append( "-outEncode は、出力ファイルのエンコードや、changeFileで指定の置換文字列" 		).append( CR );
		buf.append( "ファイルのエンコード指定になります。(changeFile は、必ず 出力ファイルと)" 		).append( CR );
		buf.append( "同じエンコードです。"															).append( CR );
		buf.append( "これらのエンコードが無指定の場合は、System.getProperty(\"file.encoding\") " 	).append( CR );
		buf.append( "で求まる値を使用します。" 														).append( CR );
		buf.append( "-changeFile を使用することで、複数行の文字列に置換することが可能です。" 		).append( CR );
		buf.append( CR );
		buf.append( "上流(プロセスチェインのデータは上流から渡されます。)からのLineModel の"		).append( CR );
		buf.append( "ファイルオブジェクトより、指定の文字列が含まれているか検索します。"			).append( CR );
		buf.append( "上流プロセスでは、Name 属性として、『File』を持ち、値は、Fileオブジェクト"		).append( CR );
		buf.append( "である、Process_FileSearch を使用するのが、便利です。それ以外のクラスを"		).append( CR );
		buf.append( "使用する場合でも、Name属性と、File オブジェクトを持つ LineModel を受け渡し"	).append( CR );
		buf.append( "できれば、使用可能です。"														).append( CR );
		buf.append( CR );
		buf.append( "引数文字列中に空白を含む場合は、ダブルコーテーション(\"\") で括って下さい。"	).append( CR );
		buf.append( "引数文字列の 『=』の前後には、空白は挟めません。必ず、-key=value の様に"		).append( CR );
		buf.append( "繋げてください。"																).append( CR );
		buf.append( CR ).append( CR );

		buf.append( getArgument().usage() ).append( CR );

		return buf.toString();
	}

	/**
	 * このクラスは、main メソッドから実行できません。
	 *
	 * @param	args	コマンド引数配列
	 */
	public static void main( final String[] args ) {
		LogWriter.log( new Process_Grep().usage() );
	}
}
